// Chad Nelson

#include "Game.h"

Game::Game() {
	snake = new Snake();
	board = new Board();
	player = new Player();
	menu = new Menu();
	cine = new Point2D(10, 10);
	difficulty = DIFFICULT;
	camera = CAMERA_FOLLOW;
	zoom = 5;
	zoom_level = 20.0;
	minimap = 0;
	highscore = 0;
	menu_in_use = GLUT_MENU_NOT_IN_USE;
}

void Game::reset() {
	highscore = max(highscore, score);
	turn = TURN_NONE;
	frame_count = 1;
	score = 0;
	lives = 3;
	player->playIntro();
	board->loadLevel(1);
	snake->setLevel(1);
	pause = false;
}

void Game::update() {
	if (menu->done_flag){
		reset();
		menu->done_flag = false;
	} else {
		if (!menu->isDone()) return;
	}

	player->updateBG();

	// Update zoom_level
	if (abs(zoom_level - zoom) > 0.01)
		zoom_level += 0.05 * ((float) zoom - zoom_level);

	if (player->isPlayingIntro()) return;
	if (player->isPlayingLevel()) return;
	if (player->isPlayingDeath()) return;

	if (menu_in_use == GLUT_MENU_IN_USE) return;

	if (pause) return;

	// Update offset of snake
	switch (difficulty) {
		case INSANE:
			snake->offset = (frame_count % 5) / 5.0 ;
			frame_count = (frame_count + 1) % 5;
			break;
		case DIFFICULT:
			snake->offset = (frame_count % 10) / 10.0 ;
			frame_count = (frame_count + 1) % 10;
			break;
		case NOVICE:
			snake->offset = (frame_count % 25) / 25.0 ;
			frame_count = (frame_count + 1) % 25;
			break;
	}
				
	// Add next block that the snake is traveling to
	if (snake->offset == 0.0) {
		snake->move();
		
		if (turn == TURN_LEFT) {
			snake->changeDirection(-1);
		} else if (turn == TURN_RIGHT) {
			snake->changeDirection(1);
		}

		turn = TURN_NONE;

		// Check for object collisions
		for (unsigned int i = 0; i < board->objects.size(); i++) {
			WallSegment *w = board->objects[i];
			if (snake->next_pos->x == w->position->x && snake->next_pos->y == w->position->y) {
				// Collision with wall!
				loseLife();
				return;
			}
		}

		// Check for boundary leaving
		if (snake->next_pos->x < 0 || snake->next_pos->x > board->width - 1 ||
			snake->next_pos->y < 0 || snake->next_pos->y > board->height - 1) {
				loseLife();
				return;
		}

		// Check for running into tail of snake
		list<Point2D *>::iterator it1;
		for(it1 = snake->snake_coordinates.begin(); it1 != snake->snake_coordinates.end(); ++it1) {
			Point2D *p = *it1;
			if (snake->next_pos->x == p->x && snake->next_pos->y == p->y) {
				// Running into tail!
				loseLife();
				return;
			}
		}

		// Check for pellet
		list<Pellet *>::iterator it2;
		for(it2 = board->pellets.begin(); it2 != board->pellets.end(); ++it2) {
			Pellet *p = *it2;
			if (snake->snake_coordinates.front()->x == p->position->x && snake->snake_coordinates.front()->y == p->position->y) {
				// Eat pellet!
				score += p->points;
				it2 = board->pellets.erase(it2);
				player->playEat();
				snake->grow += 1 + p->growth;

				// End of level check
				if (board->pellets.size() == 0) {
					score += 10 * snake->snake_coordinates.size();

					if (board->level == 5) {
						reset();
						return;
					}
						
					setLevel(board->level + 1);
				}
				return;
			}
		}
	}
}

void Game::setLevel(int l) {
	if (l < 1 || l > 5) return;
	board->loadLevel(l);
	player->playLevel();
	snake->setLevel(l);
	frame_count = 1;
}

void Game::loseLife() {
	lives--;
	player->playDeath();
	snake->setLevel(board->level);
	
	if (lives <= 0)
		reset();		
}

void Game::draw() {
	if (!menu->isDone()) return;

	glPushMatrix();

	// Update camera
	if (camera == CAMERA_TOP) {
		gluLookAt(snake->x * TILE_SIZE      , zoom_level * 20.0      , snake->y * TILE_SIZE, 
			      snake->x * TILE_SIZE      , 0.0                    , snake->y * TILE_SIZE, 
				  snake->x - snake->x_follow, 0.0                    , snake->y - snake->y_follow); 
	} else if (camera == CAMERA_FOLLOW) {
		float px = snake->x + zoom_level * (snake->x_follow - snake->x);
		float py = snake->y + zoom_level * (snake->y_follow - snake->y);

		gluLookAt(px * TILE_SIZE      , 2.0 * pow(1.2f, zoom_level) * TILE_SIZE, py * TILE_SIZE, 
			      snake->x * TILE_SIZE, 0.0                                    ,snake->y * TILE_SIZE, 
				  0.0                 , 1.0                                    , 0.0); 
	} else {
		gluLookAt(snake->x * TILE_SIZE      , zoom_level * 20.0      , snake->y * TILE_SIZE, 
			      snake->x * TILE_SIZE      , 0.0                    , snake->y * TILE_SIZE, 
				  snake->x - snake->x_follow, 0.0                    , snake->y - snake->y_follow);
	}

	// Position the sun
	float sun_origin[] = {100.0 , -100.0, 150.0};
	glLightfv(GL_LIGHT0, GL_POSITION, sun_origin);

	// Draw walls
	snake->draw();
	board->draw();
		
	// Draw reflection
	glEnable(GL_STENCIL_TEST); //Enable using the stencil buffer
	glColorMask(0, 0, 0, 0); //Disable drawing colors to the screen
	glDisable(GL_DEPTH_TEST); //Disable depth testing
	glStencilFunc(GL_ALWAYS, 1, 1); //Make the stencil test always pass
	//Make pixels in the stencil buffer be set to 1 when the stencil test passes
	glStencilOp(GL_KEEP, GL_KEEP, GL_REPLACE);
	//Set all of the pixels covered by the floor to be 1 in the stencil buffer
	board->drawEvenSquares();
	board->drawOddSquares();
	
	glColorMask(1, 1, 1, 1); //Enable drawing colors to the screen
	glEnable(GL_DEPTH_TEST); //Enable depth testing
	//Make the stencil test pass only when the pixel is 1 in the stencil buffer
	glStencilFunc(GL_EQUAL, 1, 1);
	glStencilOp(GL_KEEP, GL_KEEP, GL_KEEP); //Make the stencil buffer not change
	
	//Draw the reflected objectsy, at all pixels where the stencil buffer is 1
	glPushMatrix();
		glScalef(1, -1, 1);
		glLightfv(GL_LIGHT0, GL_POSITION, sun_origin);
		snake->draw();
		board->draw();
	glPopMatrix();
	
	glDisable(GL_STENCIL_TEST); //Disable using the stencil buffer
	
	//Blend the floor onto the screen
	glEnable(GL_BLEND);
	board->drawOddSquares();
	board->drawEvenSquares();
	glDisable(GL_BLEND);

	glPopMatrix();
}

void Game::draw2D() {
	if (!menu->isDone()) {
		menu->draw();
		return;
	}

	// Draw Score
	glEnable(GL_BLEND);
		glBegin(GL_QUADS);
			glColor4f(0, 0, 0, 0.4);
			glVertex3f(0, glutGet(GLUT_WINDOW_HEIGHT), -0.5);
			glVertex3f(0, glutGet(GLUT_WINDOW_HEIGHT) - 25, -0.5);
			glVertex3f(glutGet(GLUT_WINDOW_WIDTH), glutGet(GLUT_WINDOW_HEIGHT) - 25, -0.5);
			glVertex3f(glutGet(GLUT_WINDOW_WIDTH), glutGet(GLUT_WINDOW_HEIGHT), -0.5);
		glEnd();
	glDisable(GL_BLEND);

	char textL[256];
	char textR[256];
	sprintf_s(textL, "Score: %d  Lives: %d  High Score: %d\0", score, lives, highscore);
	sprintf_s(textR, "Level: %d of 5\0", board->level);
	drawText(8, glutGet(GLUT_WINDOW_HEIGHT) - 16, textL);
	drawText(glutGet(GLUT_WINDOW_WIDTH) - 122, glutGet(GLUT_WINDOW_HEIGHT) - 16, textR);

	
	// Draw minimap
	if (minimap > 0) {
			glPushMatrix();
				if (minimap == 1) {
					glScalef(1.0, 1.0, 1.0 / (float) TILE_SIZE);
					glRotatef(90.0, 1.0, 0.0, 0.0);
					glTranslatef(TILE_SIZE * 2, 0, -(board->height + 2) * TILE_SIZE);
				} else {
					glScalef(0.5, 0.5, 0.5 / (float) TILE_SIZE);
					glRotatef(90.0, 1.0, 0.0, 0.0);
					glTranslatef(TILE_SIZE * 3, 0, -(board->height + 3) * TILE_SIZE);
				}
				glEnable(GL_BLEND);
				board->drawEvenSquares();
				board->drawOddSquares();
				glDisable(GL_BLEND);
				board->draw();
				snake->draw();
			glPopMatrix();
	}

	// Draw paused indicator
	if (pause) {
		Point2D *center = new Point2D(glutGet(GLUT_WINDOW_WIDTH)/2, glutGet(GLUT_WINDOW_HEIGHT)/2);
		int w = 32;
		int h = 10;
		glEnable(GL_BLEND);
			glBegin(GL_QUADS);
				glColor4f(0, 0, 0, 0.4);
				glVertex3f(center->x - w, center->y - h, -0.5);
				glVertex3f(center->x + w, center->y - h, -0.5);
				glVertex3f(center->x + w, center->y + h, -0.5);
				glVertex3f(center->x - w, center->y + h, -0.5);
			glEnd();
		glDisable(GL_BLEND);
		drawText(center->x - w + 5, center->y - h + 5, "Paused\0");
	}

}

void Game::changeDifficulty(int d) {
	if (difficulty == INSANE) {
		if (d == NOVICE) {
			frame_count *= 5;
		} else if (d == DIFFICULT) {
			frame_count *= 2;
		}
	} else if (difficulty == DIFFICULT) {
		if (d == NOVICE) {
			frame_count *= 2.5;
		} else if (d == INSANE) {
			frame_count /= 2;
		}
	} else {
		if (d == DIFFICULT) {
			frame_count /= 2.5;
		} else if (d == INSANE) {
			frame_count /= 5;
		}
	}

	difficulty = d;
}


// my implementation of glutBitmapString
void Game::drawText(int x, int y, char* text) {
	for (int i=0; text[i] != '\0'; i++) {
		// Draw Text
		glColor3f(1.0f, 1.0f, 1.0f);
		glRasterPos2i(x + i * 9, y);
		glutBitmapCharacter(GLUT_BITMAP_9_BY_15, text[i]);		
	}
}